from __future__ import annotations

from typing import Any, MutableMapping, Optional, Sequence, cast, overload

from pydantic import BaseModel
from rcheck import r

from pnorm import MarshallRecordException
from pnorm.pnorm_types import MappingT, ParamType, T


def get_params(
    name: str,
    params: Optional[ParamType],
    by_alias: bool = False,
) -> dict[str, Any] | None:
    if params is None:
        return None

    if isinstance(params, BaseModel):
        params = params.model_dump(by_alias=by_alias, mode="json")

    return cast(
        dict[str, Any],
        r.check_mapping(name, params, keys_of=str, values_of=Any),
    )


@overload
def combine_into_return(
    return_model: type[T],
    result: MutableMapping[str, Any] | BaseModel,
    params: Optional[ParamType] = None,
) -> T: ...


@overload
def combine_into_return(
    return_model: type[MappingT],
    result: MutableMapping[str, Any] | BaseModel,
    params: Optional[ParamType] = None,
) -> MappingT: ...


def combine_into_return(
    return_model: type[T] | type[MappingT],
    result: MutableMapping[str, Any] | BaseModel,
    params: Optional[ParamType] = None,
) -> T | MappingT:
    result_dict = get_params("Query Result", result)

    if params is not None:
        result_dict.update(get_params("Query Params", params))

    try:
        return return_model(**result_dict)
    except Exception as e:
        model_name = getattr(return_model, "__name__")
        msg = f"Could not marshall record {result_dict} into model {model_name}"
        raise MarshallRecordException(msg) from e


@overload
def combine_many_into_return(
    return_model: type[T],
    result: Sequence[MutableMapping[str, Any] | BaseModel],
    params: Optional[ParamType] = None,
) -> tuple[T, ...]: ...


@overload
def combine_many_into_return(
    return_model: type[MappingT],
    result: Sequence[MutableMapping[str, Any] | BaseModel],
    params: Optional[ParamType] = None,
) -> tuple[MappingT, ...]: ...


def combine_many_into_return(
    return_model: type[T] | type[MappingT],
    results: Sequence[MutableMapping[str, Any] | BaseModel],
    params: Optional[ParamType] = None,
) -> tuple[T, ...] | tuple[MappingT, ...]:

    return tuple(
        combine_into_return(return_model, result, params) for result in results
    )



import os
from glob import glob
from importlib import import_module
from pathlib import Path
from typing import Any, Sequence, Type

import click

from pnorm import linter, migrations
from pnorm.linter import PostgresCredentials


@click.group()
def cli(): ...


@cli.command()
@click.option("--path", default=".", help="Path to file or directory")
def lint(path: str) -> None:
    raise NotImplementedError("linting is not currently enabled")
    expanded_path = os.path.expanduser(path)
    abs_path = os.path.abspath(expanded_path)
    linter.main(Path(abs_path))


@cli.command()
@click.option("--path", default=".", help="Path to file or directory")
@click.option("--host")
@click.option("--dbname", default=None)
@click.option("--database", default=None)
@click.option("--port", type=int)
@click.option("--user")
@click.option("--password")
def migrate(
    path: str,
    dbname: str | None,
    database: str | None,
    host: str,
    port: int,
    user: str,
    password: str,
):
    credentials = PostgresCredentials(
        dbname=dbname or database or "postgres",
        host=host,
        port=port,
        user=user,
        password=password,
    )
    migrations.main(Path(path), credentials)


if __name__ == "__main__":
    cli()



from typing import Protocol, TypedDict

from pydantic import AliasChoices, BaseModel, ConfigDict, Field, SecretStr


class CredentialsDict(TypedDict):
    dbname: str
    user: str
    password: str
    host: str
    port: int


class CredentialsProtocol(Protocol):
    dbname: str
    user: str
    password: str
    host: str
    port: int

    def as_dict(self) -> CredentialsDict: ...


class PostgresCredentials(BaseModel):
    dbname: str = Field(
        default="postgres",
        validation_alias=AliasChoices("dbname", "database"),
    )
    user: str
    password: SecretStr
    host: str
    port: int = 5432

    model_config = ConfigDict(extra="forbid")

    def as_dict(self) -> CredentialsDict:
        return {
            "dbname": self.dbname,
            "user": self.user,
            "password": self.password.get_secret_value(),
            "host": self.host,
            "port": self.port,
        }



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Generator, cast

from psycopg2._psycopg import connection as Connection
from psycopg2._psycopg import cursor as Cursor
from psycopg2.extras import RealDictCursor

from pnorm.exceptions import connection_not_created

if TYPE_CHECKING:
    from pnorm import PostgresClient


class TransactionCursor:
    def __init__(self, client: PostgresClient):
        self.client = client
        self.cursor: Cursor | None = None

    def _ensure_cursor(self) -> None:
        if self.cursor is not None:
            return

        if self.client.connection is None:
            connection_not_created()

        self.cursor = self.client.connection.cursor(cursor_factory=RealDictCursor)

    @contextmanager
    def __call__(self, _: Connection | None) -> Generator[Cursor, None, None]:
        self._ensure_cursor()

        yield cast(Cursor, self.cursor)

    def commit(self) -> None:
        if self.client.connection is None:
            connection_not_created()

        self.client.connection.commit()

    def close(self) -> None:
        self.cursor = None


class SingleCommitCursor:
    def __init__(self, client: PostgresClient):
        self.client = client

    @contextmanager
    def __call__(self, connection: Connection | None) -> Generator[Cursor, None, None]:
        if connection is None:
            connection_not_created()

        with connection:
            # with connection.cursor() as cursor:
            with connection.cursor(cursor_factory=RealDictCursor) as cursor:
                yield cursor

            connection.commit()

    def commit(self) -> None:
        if self.client.connection is None:
            connection_not_created()

        self.client.connection.commit()

    def close(self) -> None: ...



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Generator, MutableMapping, Optional, Sequence, cast, overload

import psycopg2
import psycopg2.extras as extras
from psycopg2._psycopg import connection as Connection
from psycopg2.extras import RealDictRow
from psycopg2.sql import Composable
from pydantic import BaseModel
from rcheck import r

from pnorm import (
    ConnectionAlreadyEstablishedException,
    MultipleRecordsReturnedException,
    NoRecordsReturnedException,
)
from pnorm.credentials import CredentialsDict, CredentialsProtocol, PostgresCredentials
from pnorm.cursors import SingleCommitCursor, TransactionCursor
from pnorm.exceptions import connection_not_created
from pnorm.mapping_utilities import (
    combine_into_return,
    combine_many_into_return,
    get_params,
)
from pnorm.pnorm_types import MappingT, ParamType, T


class PostgresClient:
    def __init__(
        self,
        credentials: CredentialsProtocol | CredentialsDict | PostgresCredentials,
        auto_create_connection: bool = True,
    ) -> None:
        # Want to keep as the PostgresCredentials class for SecretStr
        if isinstance(credentials, PostgresCredentials):
            self.credentials = credentials
        elif isinstance(credentials, dict):
            self.credentials = PostgresCredentials(**credentials)
        else:
            self.credentials = PostgresCredentials(**credentials.as_dict())

        self.connection: Connection | None = None
        self.auto_create_connection = r.check_bool(
            "auto_create_connection",
            auto_create_connection,
        )
        self.cursor: SingleCommitCursor | TransactionCursor = SingleCommitCursor(self)

    def set_schema(self, *, schema: str) -> None:
        schema = r.check_str("schema", schema)
        # self.execute("set search_path to %(search_path)s", {"search_path": schema})
        self.execute(f"set search_path to {schema}")

    @overload
    def get(
        self,
        return_model: type[MappingT],
        query: str | Composable,
        params: Optional[ParamType] = None,
        default: Optional[T] = None,
        combine_into_return_model: bool = False,
    ) -> MappingT: ...

    @overload
    def get(
        self,
        return_model: type[T],
        query: str | Composable,
        params: Optional[ParamType] = None,
        default: Optional[T] = None,
        combine_into_return_model: bool = False,
    ) -> T: ...

    def get(
        self,
        return_model: type[T] | type[MappingT],
        query: str | Composable,
        params: Optional[ParamType] = None,
        default: Optional[T] = None,
        combine_into_return_model: bool = False,
    ) -> T | MappingT:
        """Always returns exactly one record or raises an exception

        This method should be used by default when expecting exactly one row to
        be returned from the SQL query, such as when selecting an object by its
        unique id.

        Parameters
        ----------
        return_model : type[T of BaseModel]
            Pydantic model to marshall the SQL query results into
        query : str
            SQL query to execute
        params : Optional[Mapping[str, Any] | BaseModel] = None
            Named parameters for the SQL query
        default: T of BaseModel | None = None
            The default value to return if no rows are returned
        combine_into_return_model : bool = False
            Whether to combine the params mapping or pydantic model with the
            result of the query into the return_model

        Raises
        ------
        NoRecordsReturnedException
            When the query did not result in returning a record and no default
            was given
        MultipleRecordsReturnedException
            When the query returns at least two records

        Returns
        -------
        get : T of BaseModel
            Results of the SQL query marshalled into the return_model Pydantic model

        Examples
        --------
        >>>
        >>>
        >>>
        """
        query_params = get_params("Query Params", params)

        with self._handle_auto_connection():
            with self.cursor(self.connection) as cursor:
                cursor.execute(query, query_params)
                query_result = cast(list[RealDictRow], cursor.fetchmany(2))

        if len(query_result) >= 2:
            msg = f"Received two or more records for query: {self._query_as_string(query)}"
            raise MultipleRecordsReturnedException(msg)

        single: dict[str, Any] | BaseModel
        if len(query_result) == 0:
            if default is None:
                msg = f"Did not receive any records for query: {self._query_as_string(query)}"
                raise NoRecordsReturnedException(msg)

            single = default
        else:
            single = query_result[0]

        return combine_into_return(
            return_model,
            single,
            params if combine_into_return_model else None,
        )

    @overload
    def find(
        self,
        return_model: type[MappingT],
        query: str | Composable,
        default: MappingT,
        params: Optional[ParamType] = None,
        combine_into_return_model: bool = False,
    ) -> MappingT: ...

    @overload
    def find(
        self,
        return_model: type[T],
        query: str | Composable,
        default: T,
        params: Optional[ParamType] = None,
        combine_into_return_model: bool = False,
    ) -> T: ...

    @overload
    def find(
        self,
        return_model: type[MappingT],
        query: str | Composable,
        params: Optional[ParamType] = None,
        default: Optional[MappingT] = None,
        combine_into_return_model: bool = False,
    ) -> MappingT | None: ...

    @overload
    def find(
        self,
        return_model: type[T],
        query: str | Composable,
        params: Optional[ParamType] = None,
        default: Optional[T] = None,
        combine_into_return_model: bool = False,
    ) -> T | None: ...

    def find(
        self,
        return_model: type[T] | type[MappingT],
        query: str | Composable,
        params: Optional[ParamType] = None,
        default: Optional[T | MappingT] = None,
        combine_into_return_model: bool = False,
    ) -> T | MappingT | None:
        """Return the first result if it exists

        [desc]

        Parameters
        ----------
        return_model : type[T of BaseModel]
            Pydantic model to marshall the SQL query results into
        query : str
            SQL query to execute
        params : Optional[Mapping[str, Any] | BaseModel] = None
            Named parameters for the SQL query
        default: T of BaseModel | None = None
            The default value to return if no rows are returned
        combine_into_return_model : bool = False
            Whether to combine the params mapping or pydantic model with the
            result of the query into the return_model

        Returns
        -------
        find : T of BaseModel | None
            Results of the SQL query marshalled into the return_model Pydantic model
            or None if no rows returned

        Examples
        --------
        >>>
        >>>
        >>>
        """
        query_params = get_params("Query Params", params)
        query_result: RealDictRow | BaseModel | MappingT | None

        with self._handle_auto_connection():
            with self.cursor(self.connection) as cursor:
                cursor.execute(query, query_params)
                query_result = cast(RealDictRow | None, cursor.fetchone())

        if query_result is None or len(query_result) == 0:
            if default is None:
                return None

            query_result = default

        return combine_into_return(
            return_model,
            query_result,
            params if combine_into_return_model else None,
        )

    @overload
    def select(
        self,
        return_model: type[T],
        query: str | Composable,
        params: Optional[ParamType] = None,
    ) -> tuple[T, ...]: ...

    @overload
    def select(
        self,
        return_model: type[MappingT],
        query: str | Composable,
        params: Optional[ParamType] = None,
    ) -> tuple[MappingT, ...]: ...

    def select(
        self,
        return_model: type[T] | type[MappingT],
        query: str | Composable,
        params: Optional[ParamType] = None,
    ) -> tuple[T, ...] | tuple[MappingT, ...]:
        """Return all rows

        Parameters
        ----------
        return_model : type[T of BaseModel]
            Pydantic model to marshall the SQL query results into
        query : str
            SQL query to execute
        params : Optional[Mapping[str, Any] | BaseModel] = None
            Named parameters for the SQL query

        Returns
        -------
        select : tuple[T of BaseModel, ...]
            Results of the SQL query marshalled into the return_model Pydantic model

        Examples
        --------
        >>>
        >>>
        >>>
        """
        query_params = get_params("Query Params", params)

        with self._handle_auto_connection():
            with self.cursor(self.connection) as cursor:
                cursor.execute(query, query_params)
                query_result = cast(list[RealDictRow], cursor.fetchall())

        if len(query_result) == 0:
            return tuple()

        return combine_many_into_return(return_model, query_result)

    # todo: select using fetchmany for pagination

    def execute(
        self,
        query: str | Composable,
        params: Optional[ParamType] = None,
    ) -> None:
        """Execute a SQL query

        [desc]

        Parameters
        ----------
        query : str
            SQL query to execute
        params : Optional[Mapping[str, Any] | BaseModel] = None
            Named parameters for the SQL query

        Examples
        --------
        >>>
        >>>
        >>>
        """
        query_params = get_params("Query Params", params)

        with self._handle_auto_connection():
            with self.cursor(self.connection) as cursor:
                cursor.execute(query, query_params)

    @overload
    def execute_values(
        self,
        query: str | Composable,
        values: Optional[
            Sequence[BaseModel] | Sequence[MutableMapping[str, Any]]
        ] = None,
        *,
        template: Optional[Sequence[str]] = None,
        return_model: type[MappingT],
    ) -> tuple[MappingT, ...]: ...

    @overload
    def execute_values(
        self,
        query: str | Composable,
        values: Optional[
            Sequence[BaseModel] | Sequence[MutableMapping[str, Any]]
        ] = None,
        *,
        template: Optional[Sequence[str]] = None,
        return_model: type[T],
    ) -> tuple[T, ...]: ...

    @overload
    def execute_values(
        self,
        query: str | Composable,
        values: Optional[
            Sequence[BaseModel] | Sequence[MutableMapping[str, Any]]
        ] = None,
        *,
        template: Optional[Sequence[str]] = None,
        return_model: None = None,
    ) -> None: ...

    def execute_values(
        self,
        query: str | Composable,
        values: Optional[Sequence[BaseModel] | Sequence[MappingT]] = None,
        *,
        template: Optional[Sequence[str]] = None,
        return_model: Optional[type[T] | type[MappingT]] = None,
    ) -> tuple[T, ...] | tuple[MappingT, ...] | None:
        """Execute a sql query with values

        Parameters
        ----------
        query : str
            SQL query to execute
        values :

        template :


        Examples
        --------
        >>>
        >>>
        >>>
        """
        data: list[Any] | dict[str, Any] = []

        if values is None:
            data = get_params("Values", values)
        elif isinstance(values, list) and isinstance(values[0], tuple):
            data = values
        else:
            data = [tuple(get_params("Query params", v).values()) for v in values]

        with self._handle_auto_connection():
            with self.cursor(self.connection) as cursor:
                extras.execute_values(cursor, query, data, template)

                if return_model is None:
                    return

                query_result = cast(list[RealDictRow], cursor.fetchall())

        if len(query_result) == 0:
            return tuple()

        return combine_many_into_return(return_model, query_result)

    def _create_connection(self) -> None:
        if self.connection is not None:
            raise ConnectionAlreadyEstablishedException()

        # self.connection = psycopg2.connect(**self.credentials.as_dict())

        self.connection = psycopg2.connect(**self.credentials.as_dict())
        # self.connection = psycopg.connect(**self.credentials.as_dict(), row_factory=psycopg.rows.dict_row)

    def _end_connection(self) -> None:
        if self.connection is None:
            connection_not_created()

        self.cursor.close()
        self.connection.close()
        self.connection = None

    def _rollback(self) -> None:
        if self.connection is None:
            connection_not_created()

        self.connection.rollback()

    def _create_transaction(self) -> None:
        self.cursor = TransactionCursor(self)

    def _end_transaction(self) -> None:
        self.cursor.commit()
        self.cursor = SingleCommitCursor(self)

    @contextmanager
    def _handle_auto_connection(self) -> Generator[None, None, None]:
        close_connection_after_use = False

        if self.auto_create_connection:
            if self.connection is None:
                self._create_connection()
                close_connection_after_use = True
        elif self.connection is None:
            connection_not_created()

        try:
            yield
        finally:
            if close_connection_after_use:
                self._end_connection()

    def _query_as_string(self, query: str | Composable) -> str:
        if isinstance(query, str):
            return query

        with self._handle_auto_connection():
            with self.cursor(self.connection) as cursor:
                return query.as_string(cursor)

    @contextmanager
    def start_transaction(self) -> Generator[PostgresClient, None, None]:
        self._create_transaction()

        try:
            yield self
        except:
            self._rollback()
            raise
        finally:
            self._end_transaction()

    @contextmanager
    def start_session(
        self,
        *,
        schema: Optional[str] = None,
    ) -> Generator[PostgresClient, None, None]:
        original_auto_create_connection = self.auto_create_connection
        self.auto_create_connection = False
        close_connection_after_use = False

        if self.connection is None:
            self._create_connection()
            close_connection_after_use = True

        if schema is not None:
            self.set_schema(schema=schema)

        try:
            yield self
        except:
            self._rollback()
            raise
        finally:
            if self.connection is not None and close_connection_after_use:
                self._end_connection()

            self.auto_create_connection = original_auto_create_connection



from typing import Never


class NoRecordsReturnedException(Exception):
    """Expected to return at least one record but query did not"""


class MultipleRecordsReturnedException(Exception):
    """Expected to return at most one record but query returned multiple"""


class ConnectionAlreadyEstablishedException(Exception):
    """Trying to create a connection when one is already established. This should typically not occur"""


class ConnectionNotEstablishedException(Exception):
    """Trying to execute a query before creating an exception. Check that auto connections are turned on or you are inside a session"""


def connection_not_created() -> Never:
    """This could be from not using a session"""
    raise ConnectionNotEstablishedException()


class MarshallRecordException(Exception):
    """The returned record does not match the model you are trying to marshall it into"""



from .credentials import PostgresCredentials  # type: ignore
from .exceptions import (
    ConnectionAlreadyEstablishedException,
    ConnectionNotEstablishedException,
    MarshallRecordException,
    MultipleRecordsReturnedException,
    NoRecordsReturnedException,
)

...  # type: ignore

from .client import PostgresClient
from .model import Model, PnormConfig
from .pnorm_types import PostgresJSON

__all__ = [
    "PostgresCredentials",
    "PostgresClient",
    "Model",
    "PnormConfig",
    "NoRecordsReturnedException",
    "MultipleRecordsReturnedException",
    "ConnectionAlreadyEstablishedException",
    "ConnectionNotEstablishedException",
    "MarshallRecordException",
    "PostgresJSON",
]



import inspect
import os
import sys
from datetime import datetime
from glob import glob
from importlib import import_module, reload
from pathlib import Path
from typing import Literal

import yaml
from pydantic import BaseModel

from pnorm import PostgresClient, PostgresCredentials

from .model import Model, PnormConfig


class BaseUsers(BaseModel):
    username: str


class VersionChange(Model):
    pnorm_config: PnormConfig = PnormConfig(
        table_name="version_log",
        id_column="to_version",
    )

    from_version: int | None
    to_version: int
    update_type: Literal["upgrade", "downgrade"]
    updated_at: datetime
    description: str


class MigrationConfig(BaseModel):
    version: int
    description: str


class MigrationMetadata(BaseModel):
    path: Path
    config: MigrationConfig


class Exists(BaseModel):
    exists: bool


class VersionLogTable:

    def __init__(self, client: PostgresClient):
        self.client = client

    def handle_base_migration(self) -> None:
        if not self.table_exists:
            self.create_table()

    def get_current_version(self) -> VersionChange | None:
        return self.client.find(
            VersionChange,
            "select * from version_log order by updated_at desc limit 1",
        )

    @property
    def table_exists(self) -> bool:
        exists_check = self.client.find(
            Exists,
            "select exists ("
            "  select from information_schema.tables"
            "  where table_name = 'version_log'"
            ")",
        )

        return exists_check is not None and exists_check.exists

    def create_table(self) -> None:
        with self.client.start_transaction() as transaction:
            transaction.execute(
                "create table version_log ("
                "  from_version integer null"
                "  , to_version integer null"
                "  , description varchar"
                "  , update_type varchar not null"
                "  , updated_at timestamp not null"
                ")"
            )

            timestamp = datetime.now()

            vc = VersionChange(
                from_version=None,
                to_version=0,
                update_type="upgrade",
                updated_at=timestamp,
                description="Create version table",
            )

            vc.insert(self.client)


class Migration:
    def __init__(
        self,
        client: PostgresClient,
        version_number: int,
        version_description: str | None = None,
    ):
        self.client = client
        self.version_number = version_number
        self.version_description = version_description
        self.version_log_table = VersionLogTable(client)

    def upgrade(self, from_version: int | None) -> None:
        description = self.version_description

        if description is None:
            description = f"Upgrading from {from_version} to {self.version_number}"

        vc = VersionChange(
            from_version=from_version,
            to_version=self.version_number,
            update_type="upgrade",
            updated_at=datetime.now(),
            description=description,
        )

        vc.insert(self.client)

    def downgrade(self, to_version: int) -> None:
        vc = VersionChange(
            from_version=self.version_number,
            to_version=to_version,
            update_type="downgrade",
            updated_at=datetime.now(),
            description=f"Downgrading from {self.version_number} to {to_version}",
        )

        vc.insert(self.client)

    def run(self, current_version: VersionChange) -> None:
        if current_version.to_version > self.version_number:
            self.downgrade(current_version.to_version)
        elif current_version.to_version < self.version_number:
            self.upgrade(current_version.to_version)


def load_migration(migration: MigrationMetadata) -> type[Migration]:
    # sys.path.append(str(file_path.parent.parent))
    old_path = sys.path

    sys.path = [str(migration.path)]

    module = import_module("migration", migration.path.name)
    module = reload(module)
    classes = module.__dict__.values()

    output = None

    for cls in classes:
        if inspect.isclass(cls) and issubclass(cls, Migration) and cls != Migration:
            print("LOADING MODULE", cls.__name__, "FROM", migration.path)
            # sys.path.pop()
            # sys.path = old_path
            # return cls

            output = cls

    if output is not None:
        return output

    sys.path = old_path
    raise Exception()


def upgrade(
    session: PostgresClient,
    migrations: list[MigrationMetadata],
    previous_version: int,
    expected_version: int,
):
    versions_to_upgrade = [
        m for m in migrations if previous_version < m.config.version <= expected_version
    ]
    versions_to_upgrade = sorted(versions_to_upgrade, key=lambda x: x.config.version)

    print("VERSIONS TO UPGRADE", versions_to_upgrade)

    for version in versions_to_upgrade:
        print("from ", previous_version, "to ", version.config.version)
        migration = load_migration(version)
        migration(session, version.config.version, version.config.description).upgrade(
            previous_version
        )
        previous_version = version.config.version


def downgrade(
    session: PostgresClient,
    migrations: list[MigrationMetadata],
    previous_version: int,
    expected_version: int,
):
    versions_to_upgrade = [
        m
        for m in migrations
        if expected_version <= m.config.version <= previous_version
    ]
    versions_to_upgrade = sorted(versions_to_upgrade, key=lambda x: -x.config.version)
    print("------")
    print(versions_to_upgrade)

    for version, next_version in zip(versions_to_upgrade[:-1], versions_to_upgrade[1:]):
        migration = load_migration(version)
        migration(
            session, version.config.version, version.config.description
        ).downgrade(next_version.config.version)

    if expected_version == 0:
        version = versions_to_upgrade[-1]
        migration = load_migration(version)
        migration(
            session, version.config.version, version.config.description
        ).downgrade(0)


def main(folder_path: Path, credentials: PostgresCredentials):
    migrations: list[MigrationMetadata] = []

    with open(folder_path / "config.yaml", "r", encoding="utf-8") as file:
        file_contents = yaml.safe_load(file)

    expected_version = file_contents["version"]

    for migration_path in glob("*", root_dir=folder_path):
        if not os.path.isdir(folder_path / migration_path):
            continue

        with open(
            folder_path / migration_path / "migration.yaml",
            "r",
            encoding="utf-8",
        ) as file:
            file_contents = yaml.safe_load(file)

        migrations.append(
            MigrationMetadata(
                path=folder_path / migration_path,
                config=MigrationConfig(**file_contents),
            )
        )

    client = PostgresClient(credentials)
    with client.start_session() as session:
        version_log_table = VersionLogTable(session)
        version_log_table.handle_base_migration()

        current_version = version_log_table.get_current_version()

        if current_version is None:
            raise Exception()

    print("CURRENT VERSION", current_version, "\n\n------------")
    previous_version = current_version.to_version

    if expected_version > previous_version:
        upgrade(session, migrations, previous_version, expected_version)
    elif expected_version < previous_version:
        downgrade(session, migrations, previous_version, expected_version)



from __future__ import annotations

import json
from typing import Annotated, Any, Mapping, MutableMapping, TypeVar

from pydantic import BaseModel, PlainSerializer

T = TypeVar("T", bound=BaseModel)

QueryParams = Mapping[str, Any]
ParamType = QueryParams | BaseModel

MappingT = TypeVar("MappingT", bound=MutableMapping[str, Any])

U = TypeVar("U", dict[Any, Any] | None, list[Any] | None)

PostgresJSON = Annotated[
    U,
    PlainSerializer(json.dumps, when_used="json-unless-none"),
]



from __future__ import annotations

from contextlib import contextmanager
from datetime import datetime
from typing import (
    Any,
    Generator,
    Generic,
    Literal,
    Self,
    TypeVar,
    cast,
    get_args,
    get_origin,
    overload,
)

from psycopg2.sql import SQL, Composed, Identifier
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from rcheck import r

from pnorm.client import PostgresClient


class PnormConfig(BaseModel):
    table_name: str
    id_column: str
    parent_key_id_column: str | None = None


class ModelUtils:
    def __init__(self, model: Model | type[Model]):
        self.model = model

    @property
    def fields(self) -> dict[str, FieldInfo]:
        if not self.model.__pydantic_complete__:
            self.model.model_rebuild()

        if isinstance(self.model, Model):
            return self.model.model_fields

        return getattr(self.model, "__dict__")["model_fields"]

    @property
    def submodel_names(self) -> list[str]:
        results: list[str] = []

        for field_name, field in self.model.model_fields.items():
            if field_name != "pnorm_config" and field_sub_type(field) != "val":
                results.append(field_name)

        return results

    @property
    def submodel_names_types(self) -> list[tuple[str, str]]:
        results: list[tuple[str, str]] = []
        cls_fields = self.fields

        for submodel_name in self.submodel_names:
            results.append((submodel_name, field_sub_type(cls_fields[submodel_name])))

        return results

    @property
    def submodels(self) -> list[Model | list[Model]]:
        results: list[Model | list[Model]] = []

        for submodel_name in self.submodel_names:
            try:
                field = getattr(self.model, "__dict__")["model_fields"][submodel_name]
                results.append(field.annotation)
            except:
                results.append(getattr(self.model, submodel_name))

        return results

    @property
    def submodels_names(self) -> dict[str, Model | list[Model]]:
        results: dict[str, Model | list[Model]] = {}

        for submodel_name in self.submodel_names:
            results[submodel_name] = getattr(self.model, submodel_name)

        return results

    @property
    def non_submodel_names(self) -> list[str]:
        results: list[str] = []

        for field_name, field in self.model.model_fields.items():
            if field_name != "pnorm_config" and field_sub_type(field) == "val":
                results.append(field_name)

        return results

    def model_config(self) -> ModelConfig:
        config = self.fields["pnorm_config"].default
        return ModelConfig(
            table_name=config.table_name,
            id_column=config.id_column,
            parent_key_id_column=config.parent_key_id_column,
        )


def field_sub_type(field: FieldInfo):
    try:
        if get_origin(field.annotation) == list and issubclass(
            get_args(field.annotation)[0], Model
        ):
            return "model-list"
        elif issubclass(field.annotation, Model):
            return "model"
    except:
        return "val"

    return "val"


def sql_format(sql: str, **kwargs: Any) -> Composed:
    return SQL(sql).format(**kwargs)


class ModelConfig(BaseModel):
    table_name: str
    id_column: str
    parent_key_id_column: str | None = None


def set_col_value(column_names: list[str]) -> Composed:
    return SQL(",").join(
        [
            SQL("{col_name} = {value}").format(
                col_name=Identifier(field_name),
                value=SQL(f"%({field_name})s"),
            )
            for field_name in column_names
        ]
    )


def named_parameters(column_names: list[str]) -> Composed:
    return SQL(",").join([SQL(f"%({column_name})s") for column_name in column_names])


def col_names_as_identifier(column_names: list[str]) -> Composed:
    return SQL(",").join([Identifier(column_name) for column_name in column_names])


class Model(BaseModel):
    # TODO: can we hide in repr? (could just add underscore to pnorm_config field)
    pnorm_config: PnormConfig
    _transaction: PostgresClient | None = None

    def _get_id_value(self) -> Any:
        return getattr(self, self.pnorm_config.id_column)

    @overload
    @classmethod
    def _load_model_or_many(
        cls,
        client: PostgresClient,
        key: str,
        many: Literal[False] = False,
        use_parent: bool = False,
    ) -> Self: ...

    @overload
    @classmethod
    def _load_model_or_many(
        cls,
        client: PostgresClient,
        key: str,
        many: Literal[True] = True,
        use_parent: bool = False,
    ) -> list[Self]: ...

    @classmethod
    def _load_model_or_many(
        cls,
        client: PostgresClient,
        key: str,
        many: bool = False,
        use_parent: bool = False,
    ) -> Self | list[Self]:
        model_utils = ModelUtils(cls)
        config = model_utils.model_config()

        load_query = sql_format(
            "select * from {table_name} where {id_column} = %(id_value)s",
            table_name=Identifier(config.table_name),
            id_column=Identifier(
                config.parent_key_id_column
                if use_parent and config.parent_key_id_column is not None
                else config.id_column
            ),
        )

        if many:
            model_temp = client.select(dict[str, Any], load_query, {"id_value": key})
        else:
            model_temp = client.get(dict[str, Any], load_query, {"id_value": key})
            model_temp = [model_temp]

        cls_fields = model_utils.fields
        output: list[Self] = []

        for i, model in enumerate(model_temp):

            for submodel_name, type_ in model_utils.submodel_names_types:
                match type_:
                    case "model":
                        submodel = cast(
                            type[Model], cls_fields[submodel_name].annotation
                        )
                        model_temp[i][submodel_name] = submodel._load_model_or_many(
                            client,
                            model[config.id_column],
                            use_parent=True,
                        )
                    case "model-list":
                        submodel = cast(
                            type[Model],
                            get_args(cls_fields[submodel_name].annotation)[0],
                        )
                        model_temp[i][submodel_name] = submodel._load_model_or_many(
                            client,
                            model[config.id_column],
                            many=True,
                            use_parent=True,
                        )

            output.append(cls(**model_temp[i]))

        return output if many else output[0]

    @classmethod
    def load_model(
        cls,
        client: PostgresClient,
        key: str,
    ) -> Self:
        return cls._load_model_or_many(client, key)

    @overload
    def insert(
        self,
        transaction: PostgresClient,
        /,
        *,
        ignore_on_conflict: bool = False,
    ) -> Self: ...

    @overload
    def insert(self, /, *, ignore_on_conflict: bool = False) -> Self: ...

    def insert(
        self,
        transaction: PostgresClient | bool | None = None,
        /,
        *,
        ignore_on_conflict: bool = False,
    ) -> Self:
        if isinstance(transaction, bool):
            assert self._transaction is not None
            return self._insert(self._transaction, transaction)

        assert transaction is not None
        return self._insert(transaction, ignore_on_conflict)

    def _insert(
        self,
        transaction: PostgresClient,
        ignore_on_conflict: bool = False,
        parent_result: dict[str, Any] | None = None,
    ) -> Self:
        """Insert the model into the db if it does not otherwise exist, otherwsise do nothing"""

        # TODO: add to update too?
        if (
            parent_result is not None
            and self.pnorm_config.parent_key_id_column is not None
            and getattr(self, self.pnorm_config.parent_key_id_column) is None
        ):
            setattr(
                self,
                self.pnorm_config.parent_key_id_column,
                parent_result.get(self.pnorm_config.id_column),
            )

        output = transaction.get(
            dict[str, Any],
            sql_format(
                "insert into {table_name} ({column_names}) values ({values}) on conflict do nothing returning *",
                table_name=Identifier(self.pnorm_config.table_name),
                column_names=col_names_as_identifier(
                    self._model_utils.non_submodel_names
                ),
                values=named_parameters(self._model_utils.non_submodel_names),
            ),
            self,
        )

        for submodel_name, submodel in self._model_utils.submodels_names.items():
            match submodel:
                case Model():
                    output[submodel_name] = submodel._insert(
                        transaction,
                        ignore_on_conflict=ignore_on_conflict,
                        parent_result=output,
                    )
                case list(_):
                    sub_outputs: list[Model] = []

                    for model in submodel:
                        sub_outputs.append(
                            model._insert(
                                transaction,
                                ignore_on_conflict=ignore_on_conflict,
                                parent_result=output,
                            )
                        )

                    output[submodel_name] = sub_outputs

        return self.__class__(**output)

    def upsert(self, transaction: PostgresClient | None = None, /) -> None:
        if transaction is not None:
            return self._upsert(transaction)

        assert self._transaction is not None
        return self._upsert(self._transaction)

    def _upsert(self, transaction: PostgresClient) -> None:
        """Insert the model into the db if it does not otherwise exist, otherwise update the values"""

        transaction.execute(
            sql_format(
                "insert into {table_name} ({column_names}) values ({values}) on conflict({id_column}) do update set {set_fields}",
                table_name=Identifier(self.pnorm_config.table_name),
                column_names=col_names_as_identifier(
                    self._model_utils.non_submodel_names
                ),
                values=named_parameters(self._model_utils.non_submodel_names),
                id_column=Identifier(self.pnorm_config.id_column),
                set_fields=set_col_value(self._model_utils.non_submodel_names),
            ),
            self,
        )

        for submodel in self._model_utils.submodels:
            match submodel:
                case Model():
                    submodel.upsert(transaction)
                case list(_):
                    for model in submodel:
                        model.upsert(transaction)

    @overload
    def update_only(
        self,
        transaction: PostgresClient,
        /,
        *additional_column_names: str,
    ) -> None: ...

    @overload
    def update_only(
        self,
        first_column_name: str,
        /,
        *additional_column_names: str,
    ) -> None: ...

    def update_only(
        self,
        transaction: PostgresClient | str,
        /,
        *additional_column_names: str,
    ) -> None:
        if isinstance(transaction, str):
            assert self._transaction is not None
            return self._update_only(
                self._transaction, transaction, *additional_column_names
            )

        return self._update_only(transaction, *additional_column_names)

    def _update_only(self, transaction: PostgresClient, *column_names: str) -> None:
        column_names_to_update: list[str] = []

        for field_name in column_names:
            field_name = r.check_str(f"column-{field_name}", field_name)

            if not hasattr(self, field_name):
                raise Exception()

            match field_sub_type(getattr(self, field_name)):
                case "model":
                    submodel = cast(Model, getattr(self, field_name))
                    submodel.upsert(transaction)
                case "model-list":
                    for model in cast(list[Any], getattr(self, field_name)):
                        if not isinstance(model, Model):
                            raise Exception()

                        model.upsert(transaction)
                case "val":
                    column_names_to_update.append(field_name)

        transaction.execute(
            sql_format(
                "update {table_name} set {set_fields} where {id_col} = %(id_value)s",
                table_name=Identifier(self.pnorm_config.table_name),
                set_fields=set_col_value(column_names_to_update),
                id_col=Identifier(self.pnorm_config.id_column),
            ),
            {"id_value": self._get_id_value(), **self.model_dump(mode="json")},
        )

    def delete(self, transaction: PostgresClient | None = None, /) -> None:
        if transaction is not None:
            return self._delete(transaction)

        assert self._transaction is not None
        return self._delete(self._transaction)

    def _delete(self, transaction: PostgresClient) -> None:
        transaction.execute(
            sql_format(
                "delete from {table_name} where {id_column} = %(id_value)s",
                table_name=Identifier(self.pnorm_config.table_name),
                id_column=Identifier(self.pnorm_config.id_column),
            ),
            {"id_value": self._get_id_value()},
        )

        for submodel in self._model_utils.submodels:
            match submodel:
                case Model():
                    submodel.delete(transaction)
                case list(_):
                    for model in submodel:
                        model.delete(transaction)

    @property
    def _model_utils(self) -> ModelUtils:
        return ModelUtils(self)

    @contextmanager
    def start_transaction(
        self, client: PostgresClient, /
    ) -> Generator[Self, None, None]:
        updated_model = self.model_copy()

        try:
            # TODO: auto transaction context like in client.py
            if self._transaction is not None:
                yield updated_model
            else:
                with client.start_transaction() as transaction:
                    updated_model._set_transaction(transaction)
                    yield updated_model
        finally:
            updated_model._close_transaction()

        for field_name, value in updated_model:
            setattr(self, field_name, value)

    def _set_transaction(self, transaction: PostgresClient, /):
        self._transaction = transaction

    def _close_transaction(self):
        # TODO: or on the transaction cursor set a field to mark no longer use?
        self._transaction = None


def create_table_ddl_string(cls: Model | type[Model]) -> str:
    model_utils = ModelUtils(cls)

    datatypes = {
        str: "varchar",
        int: "int",
        float: "float",
        bool: "bool",
        datetime: "timestamp",
    }

    # previous = f'create table "{cls.pnorm_config.table_name}" (\n'
    previous = "create table [TABLE_NAME] (\n"

    for i, column_name in enumerate(model_utils.non_submodel_names):
        column_type = datatypes[model_utils.fields[column_name].annotation]

        if i == 0:
            previous += "  "
        else:
            previous += "  , "

        previous += f'"{column_name}" {column_type}\n'

    # TODO: add foreing key constraint
    # TODO: add "primary key" on id_columns
    # if cls.pnorm_config.parent_key_id_column is not None:
    #     column_name = cls.pnorm_config.parent_key_id_column
    #     # column_type = datatypes[getattr(self, column_name).annotation]
    #     column_type = datatypes[cls._model_utils.fields[column_name].annotation]
    #     previous += f'  , "{column_name}" {column_type}\n'

    previous += ");\n"

    for submodel in model_utils.submodels:
        match submodel:
            case Model():
                previous += "\n" + create_table_ddl_string(submodel)
            case list(_):
                for model in submodel:
                    previous += "\n" + create_table_ddl_string(model)
            case _:
                if issubclass(submodel, BaseModel):
                    previous += "\n" + create_table_ddl_string(submodel)

    return previous



import ast
import sys
from glob import glob
from importlib import import_module
from pathlib import Path
from typing import Any, Sequence

from pydantic import BaseModel

from pnorm import PostgresClient, PostgresCredentials

creds = PostgresCredentials(
    user="postgres",
    password="postgres",
    host="localhost",
    port=5436,
)

client = PostgresClient(creds)


def get_all_files_in_path(path: Path) -> Sequence[Path]:
    if path.is_dir():
        return tuple(path / p for p in glob("*.py", root_dir=str(path), recursive=True))

    return tuple([path])


class GetStatement(BaseModel):
    statement: str = "get"
    sql_statement: str
    return_model: type[BaseModel]


class SelectStatement(BaseModel):
    statement: str = "select"
    sql_statement: str
    return_model: type[BaseModel]


class ExecuteStatement(BaseModel):
    statement: str = "execute"
    sql_statement: str


Statement = GetStatement | SelectStatement | ExecuteStatement


def parse_statements(file_path: Path, file_contents: str) -> list[Statement]:
    tree = ast.parse(file_contents)

    statements: list[Statement] = []

    sys.path.append(str(file_path.parent.parent))
    module = import_module("test_mod.test")

    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            setattr(child, "parent", node)

        # print(ast.dump(node))

        # print(file_path)

        match node:
            case ast.Call(
                func=ast.Attribute(attr="get"),
                args=function_args,
            ):
                (return_model, sql_statement, *_) = function_args

                if not isinstance(sql_statement, ast.Constant) or not isinstance(
                    sql_statement.value, str
                ):
                    print("RAISE EXCEPTION, MUST BE HARDCODED STRING")

                statements.append(
                    GetStatement(
                        return_model=getattr(module, getattr(return_model, "id")),
                        sql_statement=getattr(sql_statement, "value"),
                    )
                )

            case ast.Call(
                func=ast.Attribute(attr="select"),
                args=function_args,
            ):
                (return_model, sql_statement, *_) = function_args

                statements.append(
                    SelectStatement(
                        sql_statement=getattr(sql_statement, "value"),
                        return_model=getattr(return_model, "id"),
                    )
                )

            case ast.Call(
                func=ast.Attribute(attr="execute"),
                args=function_args,
            ):
                (sql_statement, *_) = function_args

                statements.append(
                    ExecuteStatement(
                        # return_model=getattr(return_model, "id"),
                        sql_statement=getattr(sql_statement, "value"),
                    )
                )

    return statements


def get_params(sql: str, session: PostgresClient):
    return {"username": "alex"}


class Result(BaseModel):
    statement: Statement
    success: bool
    value: BaseModel | None = None
    exception: Any | None = None
    example_params: dict[str, str] | None = None


def execute_statements(
    statements: Sequence[Statement], session: PostgresClient
) -> list[Result]:
    results: list[Result] = []

    for statement in statements:
        match statement:
            case GetStatement(
                statement="get",
                return_model=return_model,
                sql_statement=sql_statement,
            ):
                params = get_params(sql_statement, session)
                try:
                    value = session.get(
                        return_model,
                        sql_statement,
                        params,
                    )
                    results.append(
                        Result(
                            statement=statement,
                            value=value,
                            success=True,
                            example_params=params,
                        )
                    )
                except Exception as e:
                    results.append(
                        Result(
                            statement=statement,
                            success=False,
                            exception=e,
                            example_params=params,
                        )
                    )

            case SelectStatement(
                statement="select",
                return_model=return_model,
                sql_statement=sql_statement,
            ):
                params = get_params(sql_statement, session)
                session.select(return_model, sql_statement, params)

            case ExecuteStatement(
                statement="execute",
                sql_statement=sql_statement,
            ):
                params = get_params(sql_statement, session)
                session.execute(sql_statement, params)

            case _:
                raise Exception(statement)

    return results


def lint_file(path: Path, session: PostgresClient) -> None:
    with open(path, "r", encoding="utf-8") as file:
        script = file.read()

    if script.strip() == "":
        return

    statements = parse_statements(path, script)

    results = execute_statements(statements, session)
    print(results)


def main(path: Path) -> None:
    all_files = get_all_files_in_path(path)

    with client.start_session() as session:
        for file in all_files:
            lint_file(file, session)
